from __future__ import annotations

import argparse
from enum import Enum
import os.path
import sys
from typing import (
    Callable,
    Dict,
    IO,
    Optional,
    List,
    Protocol,
    Set,
    Tuple,
    TypeVar,
)

from e3.testsuite.report.index import ReportIndex, ReportIndexEntry
from e3.testsuite.report.xunit import dump_xunit_report
from e3.testsuite.result import (
    FailureReason,
    TestResult,
    TestResultSummary,
    TestStatus,
)
from e3.testsuite.utils import ColorConfig


args_parser = argparse.ArgumentParser(
    description="Display the results of a testsuite."
)
args_parser.add_argument(
    "--force-colors",
    "-C",
    action="store_true",
    help="Force the use of colors in the output.",
)
args_parser.add_argument(
    "--all-logs",
    "-a",
    action="store_true",
    help="Show logs for all tests, even if successful.",
)
args_parser.add_argument(
    "--xfail-logs", action="store_true", help="Display the log of XFAIL tests."
)
args_parser.add_argument(
    "--show-error-output",
    "-E",
    action="store_true",
    dest="show_error_output",
    default=True,
    help="Display the log of test failures. Enabled by default.",
)
args_parser.add_argument(
    "--no-error-output",
    action="store_false",
    dest="show_error_output",
    help="Do not display test output logs.",
)
args_parser.add_argument(
    "--show-time-info",
    action="store_true",
    help="Display time information for test results, if available.",
)
args_parser.add_argument(
    "--xunit-name",
    dest="xunit_name",
    metavar="NAME",
    default="Untitled testsuite",
    help="Name to use as the XUnit report name.",
)
args_parser.add_argument(
    "--xunit-output",
    dest="xunit_output",
    metavar="FILE",
    help="Output testsuite report to the given file in the standard XUnit XML"
    " format. This is useful to display results in continuous build systems"
    " such as Jenkins.",
)
args_parser.add_argument(
    "--old-result-dir",
    help="Directory that contains the report from a previous testsuite run. If"
    " passed, used to compute the new/already-detected/fixed regressions.",
)
args_parser.add_argument(
    "report",
    metavar="RESULT_DIR",
    nargs="?",
    default=os.path.join("out", "new"),
    help="Directory that contains the report to load. By default, use"
    " 'out/new' from the current directory.",
)
args_parser.add_argument(
    "--failure-exit-code",
    metavar="N",
    type=int,
    default=0,
    help="Exit code to use when at least one test result shows a"
    " failure/error. By default, this is 0. This option is useful when running"
    " this script in a continuous integration setup, as this can make the"
    " testing process stop when there is a regression.",
)


class SupportsLessThan(Protocol):
    def __lt__(self, other: SupportsLessThan) -> bool: ...


KeyType = TypeVar("KeyType")


def sorted_counters(
    counters: Dict[KeyType, int], key: Callable[[KeyType], SupportsLessThan]
) -> List[Tuple[KeyType, int]]:
    """Filter out the set of null counters and sort them."""
    return sorted(
        ((key, count) for key, count in counters.items() if count),
        key=lambda couple: key(couple[0]),
    )


def summary_line(
    result: TestResultSummary, colors: ColorConfig, show_time_info: bool
) -> str:
    """Format a summary line to describe the ``result`` test result.

    :param colors: Proxy to introduce (or not) colors in the result.
    :param show_time_info: Whether to include timing information in the result.
    """
    if show_time_info and result.time is not None:
        seconds = int(result.time)
        time_info = "{:>02}m{:>02}s".format(seconds // 60, seconds % 60)
    else:
        time_info = ""

    line = "{}{:<8}{} {}{:>6}{} {}{}{}".format(
        result.status.color(colors),
        result.status.name,
        colors.Style.RESET_ALL,
        colors.Style.DIM,
        time_info,
        colors.Style.NORMAL,
        colors.Style.BRIGHT,
        result.test_name,
        colors.Style.NORMAL,
    )
    if result.msg:
        line += ": {}{}{}".format(
            colors.Style.DIM, result.msg, colors.Style.NORMAL
        )

    return line


def format_result_logs(
    result: TestResult,
    colors: ColorConfig,
    show_error_output: bool,
    show_time_info: bool,
) -> List[str]:
    """Return a human readable description for the ``result`` test result.

    :param colors: Proxy to introduce (or not) colors in the result.
    :param show_error_output: Whether to include the logs themselves: if False,
        only show the test name, status, message and timing info if requested.
    :param show_time_info: Whether to include timing information in the result.
    """
    first_line = summary_line(result.summary, colors, show_time_info)

    # If the caller has not requested logs, just return the summary line
    if not show_error_output:
        return [first_line]

    # To clearly isolate the first line from logs, frame it with horizontal
    # lines, indent the logs and insert empty lines before and after them.
    sep_line = "-" * 79
    lines = [sep_line, first_line, sep_line, ""]

    log = str(result.diff if result.diff else result.log)
    if log:
        lines.append(log)
    else:
        lines.append(f"{colors.Style.DIM}<all logs are empty>")
    lines.append(colors.Style.RESET_ALL)

    return lines


def generate_report(
    output_file: IO[str],
    new_index: ReportIndex,
    old_index: Optional[ReportIndex],
    colors: ColorConfig,
    show_all_logs: bool,
    show_xfail_logs: bool,
    show_error_output: bool,
    show_time_info: bool,
) -> None:
    """Generate a text report for testsuite results.

    :param output_file: Output file for the report.
    :param new_index: Testsuite results to display.
    :param old_index: Results from a previous testsuite run. If present, used
        to compute the new/already-detected/fixed regressions.
    :param colors: Color configuration for the output.
    :param show_all_logs: Whether to display logs for all testcases (successful
        tests are not displayed by default).
    :param show_xfail_logs: Whether to display logs for XFAIL results (hidden
        by default).
    :param show_error_output: Whether to display logs in test results.
    :param show_time_info: Whether to display time information for test
        results, if available.
    """
    # For each test name, list of lines for the results to include in the
    # detailed report (i.e. the part that follows the summary).
    results_display: Dict[str, List[str]] = {}

    # List of test names, in the order in which their results should be
    # displayed (failures first in the summary order, then non-failures if
    # requested).
    ordered_entries: List[str] = []

    count_results = len(new_index.entries)
    count_executed = 0
    count_failure_reasons: Dict[FailureReason, int] = {
        reason: 0 for reason in FailureReason
    }

    # Lists of results of interest (users may want to investigate them
    # further).
    new_failures: List[ReportIndexEntry] = []
    already_detected_failures: List[ReportIndexEntry] = []
    fixed_failures: List[ReportIndexEntry] = []
    xfail: List[ReportIndexEntry] = []
    xpass: List[ReportIndexEntry] = []
    verify: List[ReportIndexEntry] = []
    error: List[ReportIndexEntry] = []

    # Collect information from all entries
    for _, entry in sorted(new_index.entries.items()):
        # Since this is costly, load the TestResult from disk on demand
        result: Optional[TestResult] = None

        if entry.status != TestStatus.SKIP:
            count_executed += 1

        if entry.status == TestStatus.FAIL:
            # Account for entry in failure reason counters
            result = result or entry.load()
            for reason in result.failure_reasons:
                count_failure_reasons[reason] += 1

            # Register entry in the relevant failure category. Note that if
            # there is no old entry, we still consider that the failure is a
            # new one.
            old_entry = (
                old_index.entries.get(entry.test_name) if old_index else None
            )
            if old_entry:
                if old_entry.status == TestStatus.FAIL:
                    already_detected_failures.append(entry)
                else:
                    new_failures.append(entry)
            else:
                new_failures.append(entry)

        elif entry.status == TestStatus.XFAIL:
            xfail.append(entry)
        elif entry.status == TestStatus.XPASS:
            xpass.append(entry)
        elif entry.status == TestStatus.VERIFY:
            verify.append(entry)
        elif entry.status == TestStatus.ERROR:
            error.append(entry)

        # If this is a successful testcase that was failing in the previous
        # run, classify it as a "fixed failure".
        if old_index and entry.status == TestStatus.PASS:
            old_entry = old_index.entries.get(entry.test_name)
            if old_entry and old_entry.status == TestStatus.FAIL:
                fixed_failures.append(entry)

        # Unless they are requested, do not display results for successful
        # tests or XFAILed ones.
        if (
            entry.status
            in (TestStatus.PASS, TestStatus.XPASS, TestStatus.SKIP)
            and not show_all_logs
        ) or (
            entry.status == TestStatus.XFAIL
            and not show_all_logs
            and not show_xfail_logs
        ):
            continue

        result = result or entry.load()
        results_display[entry.test_name] = format_result_logs(
            result, colors, show_error_output, show_time_info
        )

    # Write the summary
    print(
        f"{colors.Style.BRIGHT}Summary:{colors.Style.NORMAL}\n",
        file=output_file,
    )
    print(f"  Out of {count_results} results", file=output_file)
    print(f"  {count_executed} executed (not skipped)", file=output_file)

    # Display test count for each status, but only for status that have
    # at least one test. Sort them by status value, to get consistent
    # order.

    def get_key(enum: Enum) -> SupportsLessThan:
        return enum.value

    stats = sorted_counters(new_index.status_counters, key=get_key)
    failure_reason_stats = sorted_counters(count_failure_reasons, key=get_key)

    if stats:
        # If this information is available, report new skipped tests and
        # removed tests.
        if old_index:

            def skip_entries(index: ReportIndex) -> Set[str]:
                return {
                    test_name
                    for test_name, entry in index.entries.items()
                    if entry.status == TestStatus.SKIP
                }

            new_skipped = len(
                skip_entries(new_index) - skip_entries(old_index)
            )
            removed = len(
                set(old_index.entries.keys()) - set(new_index.entries.keys())
            )

            print(f"  {new_skipped} new skipped test(s)", file=output_file)
            print(f"  {removed} removed test(s)", file=output_file)
            print("", file=output_file)

        # Report the number of results for each status (when the count is not
        # zero).
        for status, count in stats:
            show_reasons = failure_reason_stats and status == TestStatus.FAIL
            suffix = ", including:" if show_reasons else ""
            print(
                f"  {status.color(colors)}{status.name.ljust(12)}"
                f"{colors.Style.RESET_ALL} {count}{suffix}",
                file=output_file,
            )
            if show_reasons:
                for r, c in failure_reason_stats:
                    print(f"    {r.name.ljust(12)} {c}", file=output_file)
        print("", file=output_file)

        # Show the different categories of failures (new, already detected,
        # fixed) and in general results that users may want to inspect more
        # closely.
        if (
            new_failures
            or already_detected_failures
            or fixed_failures
            or xfail
            or xpass
            or verify
            or error
        ):
            print(
                "  The following results may need further investigation:",
                file=output_file,
            )

            def display_failures(
                kind: str,
                color: str,
                failures: List[ReportIndexEntry],
                display_if_empty: bool = False,
            ) -> None:
                if not display_if_empty and not failures:
                    return
                print(
                    f"  {color}{len(failures)} {kind}"
                    f"{colors.Style.RESET_ALL}:",
                    file=output_file,
                )
                for e in failures:
                    label = e.test_name
                    ordered_entries.append(label)
                    if e.msg:
                        label += (
                            f": {colors.Style.DIM}{e.msg}{colors.Style.NORMAL}"
                        )
                    print(f"    {label}", file=output_file)
                print("", file=output_file)

            # Always display the new failures as this is the category that
            # always make sense, even if there is no old index (assuming the
            # baseline is failure free).
            display_failures(
                "new failure(s)",
                colors.Fore.RED + colors.Style.BRIGHT,
                new_failures,
                display_if_empty=True,
            )
            display_failures(
                "already detected failure(s)",
                colors.Fore.RED,
                already_detected_failures,
            )
            display_failures(
                "fixed failure(s)", colors.Fore.GREEN, fixed_failures
            )
            display_failures("expected failure(s)", colors.Fore.CYAN, xfail)
            display_failures(
                "unexpected passed test(s)", colors.Fore.YELLOW, xpass
            )
            display_failures(
                "test(s) requiring additional verification",
                colors.Fore.YELLOW,
                verify,
            )
            display_failures(
                "test(s) aborted due to unknown error",
                colors.Fore.RED + colors.Style.BRIGHT,
                error,
            )

    else:
        print(
            f"  {colors.Style.DIM}<no test result>{colors.Style.NORMAL}\n",
            file=output_file,
        )

    # Display all non-failures at the end of the report
    non_failures = set(results_display) - set(ordered_entries)
    ordered_entries.extend(sorted(non_failures))

    # Finally, display results logs when relevant
    print(
        f"{colors.Style.BRIGHT}Result logs:{colors.Style.NORMAL}\n",
        file=output_file,
    )
    for test_name in ordered_entries:
        lines = results_display.get(test_name)
        if lines:
            for line in lines:
                print(line, file=output_file)
    if not results_display:
        print(
            f"{colors.Style.DIM}No relevant logs to display"
            f"{colors.Style.NORMAL}",
            file=output_file,
        )


def main(argv: Optional[List[str]] = None) -> int:
    args = args_parser.parse_args(argv)
    new_index = ReportIndex.read(args.report)
    old_index = (
        None
        if args.old_result_dir is None
        else ReportIndex.read(args.old_result_dir)
    )

    generate_report(
        output_file=sys.stdout,
        new_index=new_index,
        old_index=old_index,
        colors=ColorConfig(colors_enabled=args.force_colors or None),
        show_all_logs=args.all_logs,
        show_xfail_logs=args.xfail_logs,
        show_error_output=args.show_error_output,
        show_time_info=args.show_time_info,
    )

    if args.xunit_output:
        dump_xunit_report(args.xunit_name, new_index, args.xunit_output)

    # Return the appropriate status code: the failure status code from the
    # --failure-exit-code=N option when there is a least one testcase failure,
    # or 0.
    return args.failure_exit_code if new_index.has_failures else 0


if __name__ == "__main__":  # interactive-only
    main()



"""Internal helpers for e3.testsuite."""

import warnings
from typing import Any, Callable, List, TypeVar, cast


F = TypeVar("F", bound=Callable[..., Any])


def deprecated(stacklevel: int = 1) -> Callable[[F], F]:
    """
    Return a decorator to emit deprecation warnings.

    The function that the decorator returns emits the deprecation warning only
    the first time it is called.
    """

    def deprecated(func: F) -> F:
        triggered: List[bool] = []

        def wrapper(*args: Any, **kwargs: Any) -> Any:
            if not triggered:
                triggered.append(True)
                warnings.warn(
                    "This function is obsolete and will be removed in the"
                    " future",
                    DeprecationWarning,
                    stacklevel=stacklevel + 1,
                )
            return func(*args, **kwargs)

        return cast(F, wrapper)

    return deprecated



"""Lightweight index for test results.

Loading all test results for big testsuites can take a lot of time because of
all the YAML parsing involved. This module provides helpers to efficiently read
and write an index of test results. This index contains only test names,
statuses and messages, so it is super fast to read. From there, users can load
individual test full results only when needed.
"""

from __future__ import annotations

from dataclasses import dataclass
import json
import os.path
from typing import Dict, List, Optional, Set
import yaml

from e3.testsuite.result import (
    FailureReason,
    TestResult,
    TestResultSummary,
    TestStatus,
)


@dataclass
class ReportIndexEntry:
    """ReportIndex entry for a single test result."""

    index: ReportIndex
    summary: TestResultSummary
    filename: str

    @property
    def test_name(self) -> str:
        return self.summary.test_name

    @property
    def status(self) -> TestStatus:
        return self.summary.status

    @property
    def msg(self) -> Optional[str]:
        return self.summary.msg

    @property
    def failure_reasons(self) -> Set[FailureReason]:
        return self.summary.failure_reasons

    @property
    def time(self) -> Optional[float]:
        return self.summary.time

    @property
    def info(self) -> Dict[str, str]:
        return self.summary.info

    def load(self) -> TestResult:
        with open(
            os.path.join(self.index.results_dir, self.filename), "rb"
        ) as f:
            result = yaml.safe_load(f)
        assert self.summary == result.summary
        return result


class ReportIndex:
    """Lightweight index for test results."""

    INDEX_FILENAME = "_index.json"
    INDEX_MAGIC = "e3.testsuite.report.index.ReportIndex:1"

    def __init__(self, results_dir: str) -> None:
        """Initialize a ReportIndex instance."""
        self.results_dir = results_dir
        """Directory that contain test results (YAML files)."""

        self.entries: Dict[str, ReportIndexEntry] = {}
        """Map test names to their ReportIndexEntry instances."""

        self.status_counters = {s: 0 for s in TestStatus}
        """Number of test result for each test status."""

        self.duration: Optional[float] = None
        """
        Optional number of seconds for the total duration of the testsuite run.
        """

    def save_and_add_result(self, result: TestResult) -> None:
        """Save a test result in the results directory and add it to the index.

        :param result: Test result to save/add.
        """
        self.add_result(result.summary, result.save(self.results_dir))

    def add_result(self, result: TestResultSummary, filename: str) -> None:
        """Add an entry to this index for the given test result.

        Note that unlike ``save_and_add_result``, this does not write the
        result data in the results dir: it is up to the caller to make sure of
        that.

        :param result: Result to add.
        :param filename: Name of the file that contains test result data.
        """
        entry = ReportIndexEntry(self, result, filename)
        self.entries[result.test_name] = entry
        self.status_counters[result.status] += 1

    @classmethod
    def read(cls, results_dir: str) -> ReportIndex:
        """Read the index in the given results directory."""
        result = cls(results_dir)

        with open(os.path.join(results_dir, cls.INDEX_FILENAME)) as f:
            doc = json.load(f)

        # Pick the optional testsuite duration. Even though we now always
        # include this information (either a float or null) in the index JSON,
        # we still want to be able to read indexes from old versions of
        # e3-testsuite.
        duration = doc.get("duration")
        if duration is not None:
            result.duration = float(duration)

        # Basic sanity checking on the index file format
        assert (
            isinstance(doc, dict) and doc["magic"] == cls.INDEX_MAGIC
        ), "Invalid index file format"

        # Import all entries
        for e in doc["entries"]:
            result.add_result(
                TestResultSummary(
                    e["test_name"],
                    TestStatus[e["status"]],
                    e["msg"],
                    {FailureReason[fr] for fr in e["failure_reasons"]},
                    e["time"],
                    e["info"],
                ),
                e["filename"],
            )

        return result

    def write(self) -> None:
        """Write the index on disk."""
        # Create the JSON document to be the index file content
        entries: List[dict] = []
        doc = {
            "magic": self.INDEX_MAGIC,
            "duration": self.duration,
            "entries": entries,
        }
        for e in self.entries.values():
            entries.append(
                {
                    "test_name": e.test_name,
                    "status": e.status.name,
                    "msg": e.msg,
                    "failure_reasons": [fr.name for fr in e.failure_reasons],
                    "time": e.time,
                    "info": e.info,
                    "filename": e.filename,
                }
            )

        # Actually write the file
        with open(
            os.path.join(self.results_dir, self.INDEX_FILENAME), "w"
        ) as f:
            json.dump(doc, f)

    @property
    def has_failures(self) -> bool:
        """Return whether there is at least one FAIL/ERROR test status."""
        return (
            self.status_counters[TestStatus.FAIL] > 0
            or self.status_counters[TestStatus.ERROR] > 0
        )



from __future__ import annotations

import logging
import os
import sys

from e3.env import Env
from e3.main import Main
from e3.os.process import Run
import e3.yaml


def main() -> int:
    """Run e3-test script."""
    m = Main()

    # Ignore arguments here as they are arguments for the actual testsuite
    m.parse_args(known_args_only=True)

    # Find first the tool configuration file. Keep track of current directory
    # that will be used to select the test subset automatically.
    cwd = os.path.abspath(os.getcwd())
    root_dir = cwd
    while not os.path.isfile(os.path.join(root_dir, "e3-test.yaml")):
        new_root_dir = os.path.dirname(root_dir)
        if new_root_dir == root_dir:
            logging.error("cannot find e3-test.yaml")
            return 1
        root_dir = new_root_dir
    config_file = os.path.join(root_dir, "e3-test.yaml")

    config = e3.yaml.load_with_config(config_file, Env().to_dict())

    if "main" not in config:
        logging.error("cannot find testsuite main")
        return 1
    p = Run(
        [
            sys.executable,
            os.path.join(root_dir, config["main"]),
            os.path.relpath(cwd, root_dir) + "/",
        ]
        + config.get("default_args", []),
        output=None,
        cwd=root_dir,
    )
    assert isinstance(p.status, int)
    return p.status



"""Miscellaneous helpers."""

from __future__ import annotations

from enum import Enum, auto
import os
import sys
from typing import (
    AnyStr,
    Dict,
    IO,
    Iterator,
    Optional,
    Type,
    TypeVar,
    TYPE_CHECKING,
)

from e3.os.fs import unixpath
from e3.os.process import quote_arg

if TYPE_CHECKING:
    from e3.env import Env


def isatty(stream: IO[AnyStr]) -> bool:
    """Return whether stream is a TTY.

    This is a safe predicate: it works if stream is None or if it does not even
    support TTY detection: in these cases, be conservative (consider it's not a
    TTY).
    """
    return bool(stream) and hasattr(stream, "isatty") and stream.isatty()


class DummyColors:
    """Stub to replace colorama's Fore/Style when colors are disabled."""

    def __getattr__(self, name: str) -> str:
        return ""


class ColorConfig:
    """Proxy for color management.

    This embeds colorama's Fore/Style, or DummyColors instances when colors are
    disabled.
    """

    def __init__(self, colors_enabled: Optional[bool] = None):
        """
        Initialize a ColorConfig instance.

        :param colors: Whether to enable colors. If left to None, enable it iff
            the standard output is a TTY.
        """
        from colorama import Fore, Style

        self.Fore = Fore
        self.Style = Style

        if colors_enabled is None:
            colors_enabled = isatty(sys.stdout)

        if not colors_enabled:
            self.Fore = DummyColors()
            self.Style = DummyColors()


class CleanupMode(Enum):
    """Mode for working space cleanups."""

    NONE = auto()
    PASSING = auto()
    ALL = auto()

    @classmethod
    def default(cls) -> CleanupMode:
        return cls.PASSING

    @classmethod
    def descriptions(cls) -> Dict[CleanupMode, str]:
        return {
            cls.NONE: "Remove nothing.",
            cls.PASSING: "Remove only working spaces for passing tests"
            " (useful for post-mortem investigation with reasonable disk"
            " usage). This is the default.",
            cls.ALL: "Remove all working spaces.",
        }


EnumType = TypeVar("EnumType", bound=Enum)


def enum_to_cmdline_args_map(enum_cls: Type[EnumType]) -> Dict[str, EnumType]:
    """Turn enum alternatives into command-line arguments.

    This helps exposing enums for options on the command-line. This turns
    alternative names into lower case and replaces underscores with dashes.
    """
    return {value.name.lower().replace("_", "-"): value for value in enum_cls}


def dump_environ(filename: str, env: Env) -> None:
    """Dump environment variables into a sourceable file."""
    with open(os.path.join(filename), "w") as f:
        for var_name in sorted(os.environ):
            if (
                # Ignore environment variables whose names will make
                # the "export" commands invalid. Such variables (for
                # instance PROGRAMFILES(X86) on Windows systems) are
                # generally set system-wide, so capturing them here is
                # not useful.
                "(" in var_name
                # Also ignore variables known to be readonly on Cygwin
                # systems. Other users are unlikely to be affected.
                or var_name in ("PROFILEREAD", "SHELLOPTS")
            ):
                continue

            var_value = os.environ[var_name]

            # For Cygwin tools, turn Windows-style dirnames to
            # Unix-style ones for PATH.
            if (
                var_name == "PATH"
                and env.build.os.name == "windows"
                and os.path.pathsep in var_value
            ):
                var_value = ":".join(
                    unixpath(p) for p in var_value.split(os.path.pathsep)
                )

            f.write(f"export {var_name}={quote_arg(var_value)}\n")


def indent(text: str, prefix: str = "  ") -> str:
    """Prepend ``prefix`` to every line in ``text``.

    :param text: Text to transform.
    :param prefix: String to prepend.
    """
    # Use .split() rather than .splitlines() because we need to preserve the
    # last line if is empty. "a\n".splitlines() returns ["a"], so we must avoid
    # it.
    return "\n".join((prefix + line) for line in text.split("\n"))


def safe_dir_walk(top: str) -> Iterator[tuple[str, list[str], list[str]]]:
    """Traverse a directory hierarchy following symlinks in a safe way.

    This is essentially a wrapper around os.walk() to safely follow symbolic
    links that keeps track of the directories traversed to avoid infinite
    recursion in case of symbolic link loops.
    """
    # Set of realpaths (os.path.realpath) for already visited directories
    visited: set[str] = set()

    def already_visited(f: str) -> bool:
        """Return whether we already visited a directory.

        If ``f`` designates an already visited directory, return False.
        Otherwise, keep track of it as a visited directory and return True.
        """
        f = os.path.realpath(f)
        if f in visited:
            return True
        else:
            visited.add(f)
            return False

    def recurse(top: str) -> Iterator[tuple[str, list[str], list[str]]]:
        """Recursive wrapper around os.walk()."""
        if already_visited(top):
            return
        for dirpath, dirnames, filenames in os.walk(top):
            yield dirpath, dirnames, filenames

            # For directorise that are actually symlinks, recurse to traverse
            # them too.
            for d in dirnames:
                d = os.path.join(dirpath, d)
                if os.path.islink(d):
                    yield from recurse(d)

    yield from recurse(top)



from __future__ import annotations

"""Special scheduler for efficient use of subprocesses.

When the parallelism level is high enough, using threads to dispatch the work
to do creates too much contention on Python's GIL. Since `e3.job.scheduler` is
based on threads, this module provides an alternative API to avoid creating
threads, dispatching work using subprocesses instead.
"""

import itertools
import logging
import time
from typing import Callable, Generic, List, Optional, TYPE_CHECKING, TypeVar

from e3.collection.dag import DAGIterator
from e3.env import Env
from e3.os.process import Run

if TYPE_CHECKING:
    from e3.collection.dag import DAG
    from e3.testsuite.driver import TestDriver


logger = logging.getLogger("testsuite.process_scheduler")


class Worker:
    """Abstract class to represent units of work for the scheduler."""

    index_generator = iter(itertools.count(1))
    """Generate unique indexes for each worker."""

    def __init__(
        self,
        uid: str,
        driver: TestDriver,
        callback_name: str,
        slot: int,
        env: Env,
    ):
        """Initialize a Worker instance.

        :param uid: Unique identifier for this worker.
        :param driver: Test driver handling the work to do.
        :param callback_name: Name of the `driver` method to call for the work
            to do.
        :param slot: ID for the slot this worker was assigned to. At all time,
            all running workers are assigned different slots.
        """
        self.uid = uid
        self.driver = driver
        self.callback_name = callback_name
        self.slot = slot
        self.env = env
        self.index = next(self.index_generator)

        self.process = self.start()
        """Process that executes this test fragment."""

    def start(self) -> Run:
        """Create and return the subprocess to do the work.

        All subclasses must override this.
        """
        raise NotImplementedError

    def poll(self, scheduler: MultiprocessScheduler) -> bool:
        """Return whether the subprocess is still running.

        If it is, the caller should invoke
        `MultiprocessScheduler.collect_result` on it.
        """
        if self.process.poll() is not None:
            self.process.wait()
            return False
        else:
            return True


WorkData = TypeVar("WorkData")
"""Type that contains all the information needed to do some unit of work."""

SomeWorker = TypeVar("SomeWorker", bound=Worker)
"""Worker subclass to start some unit of work."""

JobFactoryCallback = Callable[[str, WorkData, int], SomeWorker]
"""Callback to create a Worker instance from work data.

Arguments are:

* UID for this unit of work;
* data for the work to do;
* slot ID for the new worker.

Returned value is the Worker instance.
"""

CollectResultCallback = Callable[[SomeWorker], None]
"""Callback to extract work result from a worker."""


class MultiprocessScheduler(Generic[WorkData, SomeWorker]):
    """Scheduler to dispatch units of work to subprocesses."""

    def __init__(
        self,
        dag: DAG,
        job_factory: JobFactoryCallback,
        collect_result: CollectResultCallback,
        jobs: int = 0,
        dyn_poll_interval: bool = True,
    ):
        """Initialize a MultiprocessScheduler instance.

        :param dag: DAG in which nodes represent units of work to do and edges
            represent dependencies between them.
        :param job_factory: Callback to turn DAG nodes into corresponding
            Worker instances.
        :param collect_result: Callback to extract work result from a worker.
        :param jobs: Maximum of worker allowed to run in parallel. If left to
            0, use the number of available cores on the current machine.
        :param dyn_poll_interval: If True the interval between each polling
            iteration is automatically updated. Otherwise it's set to 0.1
            seconds.
        """
        e = Env()
        self.parallelism = jobs or e.build.cpu.cores
        self.dag = dag

        self.workers: List[Optional[SomeWorker]] = [None] * self.parallelism
        """
        List of active workers. Indexes in this list correspond to slot IDs
        passed to workers: `self.workers[N].slot == N` for all present
        wor,kers. When the worker is done, we just replace it with None, and
        when a slot is None we can create a new worker for it.
        """

        self.iterator = DAGIterator(self.dag, enable_busy_state=True)
        """Iterator to get ready-to-run units of work."""

        self.job_factory = job_factory
        self.collect_result = collect_result

        self.active_workers = 0
        """Equivalent to the number of non-None slots in ``self.workers``."""

        self.poll_interval = 0.1
        """Time (in seconds) to wait between each round of worker polling."""

        self.dyn_poll_interval = dyn_poll_interval

        self.no_free_item = False
        """
        True if there is work waiting to be executed, False if all work to be
        scheduled depends on work that hasn't completed.
        """

        self.no_work_left = False
        """
        True if we processed all items from ``self.iterator`` (i.e. we got a
        ``StopIteration`` exception from it).
        """

    @property
    def has_free_slots(self) -> bool:
        """Return whether there is a free slot to spawn a worker."""
        return self.active_workers < self.parallelism

    def spawn_worker(self, uid: str, data: WorkData, slot: int) -> None:
        """Create a worker and assign it to the given slot."""
        assert self.workers[slot] is None
        worker = self.job_factory(uid, data, slot)
        self.workers[slot] = worker
        self.active_workers += 1

    def release_worker(self, slot: int) -> None:
        """Release a worker, freeing the corresponding slot."""
        assert self.workers[slot] is not None
        self.workers[slot] = None
        self.active_workers -= 1

    def run(self) -> None:
        """Run the loop to execute all units of work."""
        # Run the main loop until all fragments are started and have
        # completed. We need to wait for started fragments when catching a
        # KeybordInterrupt exception (user asked to stop, or the testsuite
        # decided to stop because of too many consecutive failures).
        try:
            while self.active_workers > 0 or not self.no_work_left:
                self.poll()

        except KeyboardInterrupt:  # interactive-only
            logger.error(
                "Scheduling abortion requested, waiting for all active"
                " workers..."
            )

            # Poll active workers at regular but small interval until they all
            # complete.
            while self.active_workers > 0:
                for slot, worker in enumerate(self.workers):
                    if worker is None:
                        continue

                    # If the worker has completed, release it, but do not
                    # collect it: we are not interested in test results created
                    # after testsuite abortion.
                    still_running = worker.poll(self)
                    if not still_running:
                        self.release_worker(slot)
                        self.iterator.leave(worker.uid)

                time.sleep(0.1)

            # Let the caller know about abnormal interruption
            raise

    def poll(self) -> None:
        # Perform a linear scan to find free slots: allocate a worker for each
        # of them.
        #
        # Note that there is no need to go through this if we already know
        # that:
        #
        # * there is no work left to schedule;
        # * all slots are occupied by workers;
        # * all pending work units depend on non-completed units.
        if (
            not self.no_work_left
            and self.has_free_slots
            and not self.no_free_item
        ):
            for slot, worker in enumerate(self.workers):
                if worker is None:
                    # Three possible cases:
                    #
                    # * At least one work unit can be scheduled right now: the
                    #   call to `next` return non-None results.
                    #
                    # * There are work units left, but they all depend on
                    #   non-completed other units: we get two None values.
                    #
                    # * All work units were scheduled: the call to `next`
                    #   raises a StopIteration exception.
                    try:
                        uid, work_data = next(self.iterator)
                    except StopIteration:
                        self.no_work_left = True
                        break

                    if work_data is None:
                        # All pending work units depend on non-completed units.
                        # There is no need to continue scanning workers.
                        self.no_free_item = True
                        break

                    assert isinstance(uid, str)
                    self.spawn_worker(uid, work_data, slot)

                    # No need to continue scanning if that was the last free
                    # slot.
                    if not self.has_free_slots:
                        break

        # Now, wait for some work units to complete if either:
        #
        # * all worker slots are busy;
        # * all pending work depends on non-completed work;
        # * there is no work left to schedule but we have some workers to wait.
        logger.debug("Wait for free worker")
        poll_counter = 0
        while (
            not self.has_free_slots
            or self.no_free_item
            or (self.no_work_left and self.active_workers > 0)
        ):
            poll_counter += 1
            for slot, worker in enumerate(self.workers):
                # If the worker has completed, release the corresponding slot
                if worker is not None and not worker.poll(self):
                    self.release_worker(slot)
                    self.iterator.leave(worker.uid)
                    self.no_free_item = False
                    self.collect_result(worker)

            time.sleep(self.poll_interval)

        # Adjust the poll interval if it is dynamic
        if self.dyn_poll_interval:
            self.poll_interval = compute_next_dyn_poll(
                poll_counter, self.poll_interval
            )


def compute_next_dyn_poll(poll_counter: int, poll_interval: float) -> float:
    """Adjust the polling interval."""
    # If we poll too often (towards busy waiting), the scheduler will waste
    # computing time. If we poll too little, we will wait for too long before
    # spawning new workers and thus we will not maximize the use of cores.
    if poll_counter > 8 and poll_interval < 1.0:
        poll_interval *= 1.25
        logger.debug(f"Increase poll interval to {poll_interval}")
    elif poll_interval > 0.0001:
        poll_interval *= 0.75
        logger.debug(f"Decrease poll interval to {poll_interval}")
    return poll_interval



"""Generic testsuite framework."""

from __future__ import annotations

import argparse
import inspect
import itertools
import logging
import os
import re
import sys
import tempfile
import time
import traceback
from typing import (
    Any,
    Callable,
    Dict,
    FrozenSet,
    IO,
    List,
    Optional,
    Pattern,
    TYPE_CHECKING,
    Tuple,
    Type,
)

from e3.collection.dag import DAG
from e3.env import Env
from e3.fs import rm, mkdir, mv
from e3.job.scheduler import Scheduler
from e3.main import Main
from e3.os.process import quote_arg
from e3.testsuite._helpers import deprecated
from e3.testsuite.report.gaia import (
    GAIAResultFiles,
    dump_gaia_report,
    dump_result_logs_if_needed,
)
from e3.testsuite.report.display import generate_report, summary_line
from e3.testsuite.report.index import ReportIndex
from e3.testsuite.report.xunit import dump_xunit_report
from e3.testsuite.result import Log, TestResult, TestStatus
from e3.testsuite.running_status import RunningStatus
from e3.testsuite.testcase_finder import (
    ParsedTest,
    ProbingError,
    TestFinder,
    YAMLTestFinder,
)
from e3.testsuite.utils import (
    CleanupMode,
    ColorConfig,
    dump_environ,
    enum_to_cmdline_args_map,
    isatty,
    safe_dir_walk,
)


if TYPE_CHECKING:
    from e3.testsuite.driver import ResultQueueItem, TestDriver
    from e3.testsuite.fragment import TestFragment


logger = logging.getLogger("testsuite")


class TestAbort(Exception):
    """Raise this to abort silently the execution of a test fragment."""

    pass


class TestsuiteCore:
    """Testsuite Core driver.

    This class is the base of Testsuite class and should not be instantiated.
    It's not recommended to override any of the functions declared in it.

    See documentation of Testsuite class for overridable methods and
    variables.
    """

    def __init__(
        self,
        root_dir: Optional[str] = None,
        testsuite_name: str = "Untitled testsuite",
    ) -> None:
        """Testsuite constructor.

        :param root_dir: Root directory for the testsuite. If left to None, use
            the directory containing the Python module that created self's
            class.
        :param testsuite_name: Name for this testsuite. It can be used to
            provide a title in some report formats.
        """
        if root_dir is None:
            root_dir = os.path.dirname(inspect.getfile(type(self)))

        self.root_dir = os.path.abspath(root_dir)
        """
        Root directory for the testsuite, i.e. directory from which the test
        directory (see ``self.test_dir``) is looked up.
        """

        self.test_dir = os.path.join(self.root_dir, self.tests_subdir)
        """
        Root directory for the tree in which testcases are searched.
        """

        logger.debug("Test directory: %s", self.test_dir)
        self.return_values: Dict[str, Any] = {}
        self.result_tracebacks: Dict[str, List[str]] = {}
        self.testsuite_name = testsuite_name

        self.running_status: RunningStatus
        """
        Object to report testsuite execution status to users and to manage
        abortion in case there are too many failures.
        """

        self.use_multiprocessing = False
        """Whether to use multi-processing for tests parallelism.

        Beyond a certain level of parallelism, Python's GIL contention is too
        high to benefit from more processors. When we reach this level, it is
        more interesting to use multiple processes to cancel the GIL
        contention.

        The actual value for this attribute is computed once the DAG is built,
        in the "compute_use_multiprocessing" method.
        """

    # Mypy does not support decorators on properties, so keep the actual
    # implementations for deprecated properties in methods.

    @deprecated(2)
    def _test_counter(self) -> int:
        return len(self.report_index.entries)

    @deprecated(2)
    def _test_status_counters(self) -> Dict[TestStatus, int]:
        return self.report_index.status_counters

    @deprecated(2)
    def _results(self) -> Dict[str, TestStatus]:
        return {
            e.test_name: e.status for e in self.report_index.entries.values()
        }

    @property
    def test_counter(self) -> int:
        """Return the number of test results in the report.

        Warning: this method is obsolete and will be removed in the future.
        """
        return self._test_counter()

    @property
    def test_status_counters(self) -> Dict[TestStatus, int]:
        """Return test result counts per test status.

        Warning: this method is obsolete and will be removed in the future.
        """
        return self._test_status_counters()

    @property
    def results(self) -> Dict[str, TestStatus]:
        """Return a mapping from test names to results.

        Warning: this method is obsolete and will be removed in the future.
        """
        return self._results()

    def compute_use_multiprocessing(self) -> bool:
        """Return whether to use multi-processing for tests parallelism.

        See docstring for the "use_multiprocessing" attribute. Subclasses are
        free to override this to take control of when multiprocessing is
        enabled. Note that this will disregard the "--force-multiprocessing"
        command line option.
        """
        raise NotImplementedError

    def testsuite_main(self, args: Optional[List[str]] = None) -> int:
        """Main for the main testsuite script.

        :param args: Command line arguments. If None, use `sys.argv`.
        :return: The testsuite status code (0 for success, a positive for
            failure).
        """
        start_time = time.time()

        self.main = Main(platform_args=True)

        # Add common options
        parser = self.main.argument_parser

        temp_group = parser.add_argument_group(
            title="temporaries handling arguments"
        )
        temp_group.add_argument(
            "-t", "--temp-dir", metavar="DIR", default=Env().tmp_dir
        )
        temp_group.add_argument(
            "--no-random-temp-subdir",
            dest="random_temp_subdir",
            action="store_false",
            help="Disable the creation of a random subdirectory in the"
            " temporary directory. Use this when you know that you have"
            " exclusive access to the temporary directory (needed in order to"
            " avoid name clashes there) to get a deterministic path for"
            " testsuite temporaries.",
        )
        temp_group.add_argument(
            "-d",
            "--dev-temp",
            metavar="DIR",
            nargs="?",
            default=None,
            const="tmp",
            help="Convenience shortcut for dev setups: forces `-t DIR"
            " --no-random-temp-subdir --cleanup-mode=none` and cleans up `DIR`"
            ' first. If no directory is provided, use the local "tmp"'
            " directory.",
        )

        cleanup_mode_map = enum_to_cmdline_args_map(CleanupMode)
        temp_group.add_argument(
            "--cleanup-mode",
            choices=list(cleanup_mode_map),
            help="Control the cleanup of working spaces.\n"
            + "\n".join(
                f"{name}: {CleanupMode.descriptions()[value]}"
                for name, value in cleanup_mode_map.items()
            ),
        )
        temp_group.add_argument(
            "--disable-cleanup",
            action="store_true",
            help="Disable cleanup of working spaces. This option is deprecated"
            " and will disappear in a future version of e3-testsuite. Please"
            " use --cleanup-mode instead.",
        )

        output_group = parser.add_argument_group(
            title="results output arguments"
        )
        output_group.add_argument(
            "-o",
            "--output-dir",
            metavar="DIR",
            default="./out",
            help="Select the output directory, where test results are to be"
            " stored (default: './out'). If --old-output-dir=DIR2 is passed,"
            " the new results are stored in DIR while DIR2 contains results"
            " from a previous run. Otherwise, the new results are stored in"
            " DIR/new/ while the old ones are stored in DIR/old. In both"
            " cases, the testsuite cleans the directory for new results"
            " first.",
        )
        output_group.add_argument(
            "--old-output-dir",
            metavar="DIR",
            help="Select the old output directory, for baseline comparison."
            " See --output-dir.",
        )
        output_group.add_argument(
            "--rotate-output-dirs",
            default=False,
            action="store_true",
            help="Rotate testsuite results: move the new results directory to"
            " the old results one before running testcases (this removes the"
            " old results directory first). If not passed, we just remove the"
            " new results directory before running testcases (i.e. just ignore"
            " the old results directory).",
        )
        output_group.add_argument(
            "--show-error-output",
            "-E",
            action="store_true",
            help="When testcases fail, display their output. This is for"
            " convenience for interactive use.",
        )
        output_group.add_argument(
            "--show-time-info",
            action="store_true",
            help="Display time information for test results, if available",
        )
        output_group.add_argument(
            "--xunit-output",
            dest="xunit_output",
            metavar="FILE",
            help="Output testsuite report to the given file in the standard"
            " XUnit XML format. This is useful to display results in"
            " continuous build systems such as Jenkins.",
        )
        output_group.add_argument(
            "--gaia-output",
            action="store_true",
            help="Output a GAIA-compatible testsuite report next to the YAML"
            " report.",
        )
        output_group.add_argument(
            "--status-update-interval",
            default=1.0,
            type=float,
            help="Minimum number of seconds between status file updates. The"
            " more often we update this file, the more often one will read"
            " garbage.",
        )

        auto_gen_default = (
            "enabled" if self.auto_generate_text_report else "disabled"
        )
        output_group.add_argument(
            "--generate-text-report",
            action="store_true",
            dest="generate_text_report",
            default=self.auto_generate_text_report,
            help=(
                f"When the testsuite completes, generate a 'report' text file"
                f" in the output directory ({auto_gen_default} by default)."
            ),
        )
        output_group.add_argument(
            "--no-generate-text-report",
            action="store_false",
            dest="generate_text_report",
            help="Disable the generation of a 'report' text file (see"
            "--generate-text-report).",
        )

        output_group.add_argument(
            "--truncate-logs",
            "-T",
            metavar="N",
            type=int,
            default=200,
            help="When outputs (for instance subprocess outputs) exceed 2*N"
            " lines, only include the first and last N lines in logs. This is"
            " necessary when storage for testsuite results have size limits,"
            " and the useful information is generally either at the beginning"
            " or the end of such outputs. If 0, never truncate logs.",
        )
        output_group.add_argument(
            "--dump-environ",
            dest="dump_environ",
            action="store_true",
            default=False,
            help="Dump all environment variables in a file named environ.sh,"
            " located in the output directory (see --output-dir). This"
            " file can then be sourced from a Bourne shell to recreate"
            " the environement that existed when this testsuite was run"
            " to produce a given testsuite report.",
        )

        exec_group = parser.add_argument_group(
            title="execution control arguments"
        )
        exec_group.add_argument(
            "--max-consecutive-failures",
            "-M",
            metavar="N",
            type=int,
            default=self.default_max_consecutive_failures,
            help="Number of test failures (FAIL or ERROR) that trigger the"
            " abortion of the testuite. If zero, this behavior is disabled. In"
            " some cases, aborting the testsuite when there are just too many"
            " failures saves time and costs: the software to test/environment"
            " is too broken, there is no point to continue running the"
            " testsuite.",
        )
        exec_group.add_argument(
            "-j",
            "--jobs",
            dest="jobs",
            type=int,
            metavar="N",
            default=Env().build.cpu.cores,
            help="Specify the number of jobs to run simultaneously",
        )
        exec_group.add_argument(
            "--failure-exit-code",
            metavar="N",
            type=int,
            default=self.default_failure_exit_code,
            help="Exit code the testsuite must use when at least one test"
            " result shows a failure/error. By default, this is"
            f" {self.default_failure_exit_code}. This option is useful when"
            " running a testsuite in a continuous integration setup, as this"
            " can make the testing process stop when there is a regression.",
        )
        exec_group.add_argument(
            "--force-multiprocessing",
            action="store_true",
            help="Force the use of subprocesses to execute tests, for"
            " debugging purposes. This is normally automatically enabled when"
            " both the level of requested parallelism is high enough (to make"
            " it profitable regarding the contention of Python's GIL) and no"
            " test fragment has dependencies on other fragments. This flag"
            " forces the use of multiprocessing even if any of these two"
            " conditions is false.",
        )
        exec_group.add_argument(
            "--skip-passed",
            action="store_true",
            help="Run only tests that did not pass in the previous testsuite"
            " run. This attempts to find successful tests in the report stored"
            " at the location where this run will create the testsuite report.",
        )
        parser.add_argument(
            "sublist", metavar="tests", nargs="*", default=[], help="test"
        )
        # Add user defined options
        self.add_options(parser)

        # Parse options
        self.main.parse_args(args)
        assert self.main.args is not None

        # If there is a chance for the logging to end up in a non-tty stream,
        # disable colors. If not, be user-friendly and automatically show error
        # outputs.
        if (
            self.main.args.log_file
            or not isatty(sys.stdout)
            or not isatty(sys.stderr)
        ):
            enable_colors = False
        else:  # interactive-only
            enable_colors = True
            self.main.args.show_error_output = True

        # If explicitly requested, disable colors
        if self.main.args.nocolor:
            enable_colors = False

        self.colors = ColorConfig(enable_colors)
        self.Fore = self.colors.Fore
        self.Style = self.colors.Style

        self.env = Env()
        self.env.enable_colors = enable_colors
        self.env.root_dir = self.root_dir
        self.env.test_dir = self.test_dir

        # Setup output directories and create an index for the results we are
        # going to produce.
        self.output_dir: str
        self.old_output_dir: Optional[str]
        self.old_report_index: ReportIndex
        self.setup_result_dirs()
        self.report_index = ReportIndex(self.output_dir)

        # Set the cleanup mode from command-line arguments
        if self.main.args.cleanup_mode is not None:
            self.env.cleanup_mode = cleanup_mode_map[
                self.main.args.cleanup_mode
            ]
        elif self.main.args.disable_cleanup:
            logger.warning(
                "--disable-cleanup is deprecated and will disappear in a"
                " future version of e3-testsuite. Please use --cleanup-mode"
                " instead."
            )
            self.env.cleanup_mode = CleanupMode.NONE
        else:
            self.env.cleanup_mode = CleanupMode.default()

        # Settings for temporary directory creation
        temp_dir: str = self.main.args.temp_dir
        random_temp_subdir: bool = self.main.args.random_temp_subdir

        # The "--dev-temp" option forces several settings
        if self.main.args.dev_temp:
            self.env.cleanup_mode = CleanupMode.NONE
            temp_dir = self.main.args.dev_temp
            random_temp_subdir = False

        # Now actually setup the temporary directory: make sure we start from a
        # clean directory if we use a deterministic directory.
        #
        # Note that we do make sure that working_dir is an absolute path, as we
        # are likely to be changing directories when running each test. A
        # relative path would no longer work under those circumstances.
        temp_dir = os.path.abspath(temp_dir)
        if not random_temp_subdir:
            self.working_dir = temp_dir
            rm(self.working_dir, recursive=True)
            mkdir(self.working_dir)

        elif not os.path.isdir(temp_dir):
            # If the temp dir is supposed to be randomized, we need to create a
            # subdirectory, so check that the parent directory exists first.
            logger.critical("temp dir '%s' does not exist", temp_dir)
            return 1

        else:
            self.working_dir = tempfile.mkdtemp("", "tmp", temp_dir)

        # Create the exchange directory (to exchange data between the testsuite
        # main and the subprocesses running test fragments). Compute the name
        # of the file to pass environment data to subprocesses.
        self.exchange_dir = os.path.join(self.working_dir, "exchange")
        self.env_filename = os.path.join(self.exchange_dir, "_env.bin")
        mkdir(self.exchange_dir)

        # Make them both available to test fragments
        self.env.exchange_dir = self.exchange_dir
        self.env.env_filename = self.env_filename

        self.gaia_result_files: Dict[str, GAIAResultFiles] = {}
        """Mapping from test names to files for results in the GAIA report."""

        # Store in global env: target information and common paths
        self.env.output_dir = self.output_dir
        self.env.working_dir = self.working_dir
        self.env.options = self.main.args

        self.running_status = RunningStatus(
            os.path.join(self.output_dir, "status"),
            self.main.args.status_update_interval,
            self.main.args.max_consecutive_failures,
        )

        # User specific startup
        self.set_up()

        # Retrieve the list of test
        self.test_list = self.get_test_list(self.main.args.sublist)

        # If requested, filter out tests that passed the previous time
        if self.main.args.skip_passed:

            def should_skip(pt: ParsedTest) -> bool:
                """Return whether the test ``pt`` should be skipped.

                It should be skipped if 1) it was run last time and 2) it
                passed or expectedly failed.
                """
                entry = self.old_report_index.entries.get(pt.test_name)
                return entry is not None and entry.status in (
                    TestStatus.PASS,
                    TestStatus.XFAIL,
                    TestStatus.XPASS,
                )

            self.test_list = [
                pt for pt in self.test_list if not should_skip(pt)
            ]

        # Create a DAG to constraint the test execution order
        dag = DAG()
        for parsed_test in self.test_list:
            self.add_test(dag, parsed_test)
        self.adjust_dag_dependencies(dag)
        dag.check()
        self.running_status.set_dag(dag)

        # Determine whether to use multiple processes for fragment execution
        # parallelism.
        self.use_multiprocessing = self.compute_use_multiprocessing()
        self.env.use_multiprocessing = self.use_multiprocessing

        # Record modules lookup path, including for the file corresponding to
        # the __main__ module.  Subprocesses will need it to have access to the
        # same modules.
        self.env.modules_search_path = list(sys.path)
        main_module_file = sys.modules["__main__"].__file__
        if main_module_file is not None:
            self.env.modules_search_path.insert(
                0, os.path.dirname(os.path.abspath(main_module_file))
            )

        # Now that the env is supposed to be complete, dump it for the test
        # fragments to pick it up.
        self.env.store(self.env_filename)

        # For debugging purposes, dump the final DAG to a DOT file
        with open(os.path.join(self.output_dir, "tests.dot"), "w") as fd:
            fd.write(dag.as_dot())

        if self.use_multiprocessing:
            self.run_multiprocess_mainloop(dag)
        else:
            self.run_standard_mainloop(dag)

        # Compute the duration for this testsuite run
        end_time = time.time()
        self.report_index.duration = end_time - start_time

        # Dump the testsuite report in all the requested formats
        self.report_index.write()
        self.dump_testsuite_result()
        if self.main.args.xunit_output:
            dump_xunit_report(
                self.testsuite_name,
                self.report_index,
                self.main.args.xunit_output,
            )
        if self.main.args.gaia_output:
            dump_gaia_report(
                report_index=self.report_index,
                output_dir=self.output_dir,
                discs=getattr(self.env, "discs", None),
                result_files=self.gaia_result_files,
            )

        # Clean everything
        self.tear_down()

        # If requested, generate a text report
        if self.main.args.generate_text_report:
            # Use the previous testsuite results for comparison, if available
            old_index = (
                ReportIndex.read(self.old_output_dir)
                if self.old_output_dir
                else None
            )

            # Include all information, except logs for successful tests, which
            # is just too verbose.
            with open(
                os.path.join(self.output_dir, "report"), "w", encoding="utf-8"
            ) as f:
                generate_report(
                    output_file=f,
                    new_index=self.report_index,
                    old_index=old_index,
                    colors=ColorConfig(colors_enabled=False),
                    show_all_logs=False,
                    show_xfail_logs=True,
                    show_error_output=True,
                    show_time_info=True,
                )

        # Return the appropriate status code: the failure status code from the
        # --failure-exit-code=N option when there is a least one testcase
        # failure, or 0.
        return (
            self.main.args.failure_exit_code
            if self.report_index.has_failures
            else 0
        )

    def get_test_list(self, sublist: List[str]) -> List[ParsedTest]:
        """Retrieve the list of tests.

        :param sublist: A list of tests scenarios or patterns.
        """
        # Use a mapping: test name -> ParsedTest when building the result, as
        # several patterns in "sublist" may yield the same testcase.
        testcases: Dict[str, ParsedTest] = {}
        test_finders = self.test_finders

        def matches_pattern(
            pattern: Optional[Pattern[str]], name: str
        ) -> bool:
            return pattern is None or bool(pattern.search(name))

        def add_testcase(
            pattern: Optional[Pattern[str]], test: ParsedTest
        ) -> None:
            # Do not add this testcase if its test-specific matcher does not
            # match the requested pattern.
            if test.test_matcher and not matches_pattern(
                pattern, test.test_matcher
            ):
                return

            if test.test_name in testcases:
                self.add_test_error(
                    test_name=test.test_name,
                    message=f"duplicate test name: {test.test_name}",
                    tb=None,
                )
            else:
                testcases[test.test_name] = test

        def helper(spec: str) -> None:
            pattern: Optional[Pattern[str]] = None

            # If the given pattern is a directory, do not go through the whole
            # tests subdirectory.
            if os.path.isdir(spec):
                root = spec
            else:
                root = self.test_dir
                try:
                    pattern = re.compile(spec)
                except re.error as exc:
                    logger.debug(
                        "Test pattern is not a valid regexp, try to match it"
                        " as-is: {}".format(exc)
                    )
                    pattern = re.compile(re.escape(spec))

            # For each directory in the requested subdir, ask our test finders
            # to probe for a testcase. Register matches.
            for dirpath, dirnames, filenames in safe_dir_walk(root):
                # Don't descend into internal VCS directories, because it will
                # likely generate a lot of unnecessary I/O operations and we
                # don't expect to find any tests there anyway.
                for vcsdir in [".git", ".svn", "CVS"]:
                    if vcsdir in dirnames:
                        dirnames.remove(vcsdir)

                # Compute whether the pattern matches the directory only once
                # (now) instead of once per test finder (in the for loop
                # below).
                pattern_matches = matches_pattern(pattern, dirpath)

                # The first test finder that has a match "wins". When handling
                # test data, we want to deal only with absolute paths, so get
                # the absolute name now.
                dirpath = os.path.abspath(dirpath)
                for tf in test_finders:
                    # If this test finder guarantees that each testcase has a
                    # dedicated directory, do not process this directory if it
                    # does not match the requested pattern.
                    if tf.test_dedicated_directory and not pattern_matches:
                        continue

                    try:
                        test_or_list = tf.probe(
                            self, dirpath, dirnames, filenames
                        )
                    except ProbingError as exc:
                        self.add_test_error(
                            test_name=self.test_name(dirpath),
                            message=str(exc),
                            tb=traceback.format_exc(),
                        )
                        break
                    if isinstance(test_or_list, list):
                        for t in test_or_list:
                            add_testcase(pattern, t)
                        break
                    elif test_or_list is not None:
                        add_testcase(pattern, test_or_list)
                        break

        # If specific tests are requested, only look for them. Otherwise, just
        # look in the tests subdirectory.
        if sublist:
            for s in sublist:
                helper(s)
        else:
            helper(self.test_dir)

        result = list(testcases.values())
        logger.info("Found {} tests".format(len(result)))
        logger.debug("tests:\n  " + "\n  ".join(t.test_dir for t in result))
        return result

    def add_test(self, dag: DAG, parsed_test: ParsedTest) -> None:
        """Register a test to run.

        :param dag: The DAG of test fragments to execute for the testsuite.
        :param parsed_test: Test to instantiate.
        """
        test_name = parsed_test.test_name

        # Complete the test environment
        test_env = dict(parsed_test.test_env)
        test_env["test_dir"] = parsed_test.test_dir
        test_env["test_name"] = test_name

        assert isinstance(self.env.working_dir, str)
        test_env["working_dir"] = os.path.join(self.env.working_dir, test_name)

        # Fetch the test driver to use
        driver = parsed_test.driver_cls
        if not driver:
            if self.default_driver:
                driver = self.test_driver_map[self.default_driver]
            else:
                self.add_test_error(
                    test_name=test_name,
                    message="missing test driver",
                )
                return

        # Finally run the driver instantiation
        try:
            instance = driver(self.env, test_env)
            instance.Fore = self.Fore
            instance.Style = self.Style
            instance.add_test(dag)

        except Exception as e:
            self.add_test_error(
                test_name=test_name,
                message=str(e),
                tb=traceback.format_exc(),
            )

    def dump_testsuite_result(self) -> None:
        """Log a summary of test results.

        Subclasses are free to override this to do whatever is suitable for
        them.
        """
        lines = ["Summary:"]

        # Display test count for each status, but only for status that have
        # at least one test. Sort them by status value, to get consistent
        # order.
        def sort_key(couple: Tuple[TestStatus, int]) -> Any:
            status, _ = couple
            return status.value

        stats = sorted(
            (
                (status, count)
                for status, count in self.report_index.status_counters.items()
                if count
            ),
            key=sort_key,
        )
        for status, count in stats:
            lines.append(
                "  {}{: <12}{} {}".format(
                    status.color(self.colors),
                    status.name,
                    self.Style.RESET_ALL,
                    count,
                )
            )
        if not stats:
            lines.append("  <no test result>")
        logger.info("\n".join(lines))

        # Dump the comment file
        with open(os.path.join(self.output_dir, "comment"), "w") as f:
            self.write_comment_file(f)

    def collect_result(self, fragment: TestFragment) -> None:
        """Import test results from ``fragment`` into testsuite reports.

        :param fragment: Test fragment (just completed) from which to import
            test results.
        """
        assert self.main.args

        # Process all results from this fragment
        while fragment.result_queue:
            self.add_result(fragment.result_queue.pop())

        # Now that this fragment is completed, make sure to remove all
        # references to its test drivers so that it can be garbage collected.
        # This is necessary to keep memory consumption under control for big
        # testsuites.
        dag = self.running_status.dag
        assert dag is not None
        dag.vertex_data[fragment.uid].clear_driver_data()
        fragment.clear_driver_data()

    def add_result(self, item: ResultQueueItem) -> None:
        """Add a test result to the result index and log it.

        :param item: Result queue item for the result to add.
        """
        assert self.main.args

        status = item.result.status
        test_name = item.result.test_name

        # The test results that reach this point are special: there were
        # serialized/deserialized through YAML, so the Log layer disappeared.
        assert status is not None

        # Ensure that we don't have two results with the same test name: if
        # that is the case, we cannot publish the result we were given: still,
        # push a "synthetic" ERROR result to keep track of the error (this is a
        # testsuite framework error).
        #
        # There is one exception to this: ERROR results are already the symptom
        # of something that went wrong in the testing framework. Because of
        # e3-testsuite's architecture (in particular multiprocessing), it is
        # not possible to have both human-friendly result names (that cleanly
        # map to test names) and reliably creating non-conflicting result names
        # at the time the test result is produced. So if the name of an ERROR
        # result conflicts with an existing result, just look for an
        # alternative name.

        def indented_tb(tb: List[str]) -> str:
            return "".join("  {}".format(line) for line in tb)

        if test_name in self.report_index.entries:
            test_name_radix = test_name
            for i in itertools.count(1):
                test_name = f"{test_name_radix}__except{i}"
                if test_name not in self.report_index.entries:
                    break

            if item.result.status != TestStatus.ERROR:
                self.add_test_error(
                    test_name,
                    f"cannot push twice results for {test_name_radix}",
                    f"\nFirst push happened at:"
                    f"\n{indented_tb(self.result_tracebacks[test_name_radix])}"
                    f"\nThis one happened at:"
                    f"\n{indented_tb(item.traceback)}",
                )
                return

        # Now that the result is validated, add it to our internals
        self.report_index.add_result(item.result, item.filename)
        self.result_tracebacks[test_name] = item.traceback
        self.running_status.process_result(item.result)
        if item.gaia_results:
            self.gaia_result_files[test_name] = item.gaia_results

        # Log the test result. If error output is requested and the test
        # failed unexpectedly, show the detailed logs.
        log_line = summary_line(
            item.result, self.colors, self.main.args.show_time_info
        )
        if self.main.args.show_error_output and status not in (
            TestStatus.PASS,
            TestStatus.XFAIL,
            TestStatus.XPASS,
            TestStatus.SKIP,
        ):
            full_result = self.report_index.entries[test_name].load()

            def format_log(log: Log) -> str:
                return "\n" + str(log) + self.Style.RESET_ALL

            if full_result.diff:
                log_line += format_log(full_result.diff)
            else:
                log_line += format_log(full_result.log)
        logger.info(log_line)

    def add_test_error(
        self, test_name: str, message: str, tb: Optional[str] = None
    ) -> None:
        """Create and add an ERROR test status.

        :param test_name: Prefix for the test result to create. This adds a
            suffix to avoid clashes.
        :param str message: Error message.
        :param tb: Optional traceback for the error.
        """
        from e3.testsuite.driver import ResultQueueItem

        result = TestResult(
            test_name, env={}, status=TestStatus.ERROR, msg=message
        )
        if tb:
            result.log += tb

        self.add_result(
            ResultQueueItem(
                result.summary,
                result.save(self.output_dir),
                traceback.format_stack(),
                dump_result_logs_if_needed(self.env, result, self.output_dir),
            )
        )

    def setup_result_dirs(self) -> None:
        """Create the output directory in which the results are stored."""
        assert self.main.args
        args = self.main.args

        # Both the actual new/old directories to use depend on both
        # --output-dir and --old-output-dir options.
        d = os.path.abspath(args.output_dir)
        if args.old_output_dir:
            self.output_dir = d
            old_output_dir = os.path.abspath(args.old_output_dir)
        else:
            self.output_dir = os.path.join(d, "new")
            old_output_dir = os.path.join(d, "old")

        # If this testsuite run should skip tests that passed in a previous
        # testsuite run, try to load the previous testsuite report.
        if args.skip_passed:
            try:
                self.old_report_index = ReportIndex.read(self.output_dir)
            except OSError as exc:
                logger.warning(
                    f"Could not load the previous testsuite report: {exc}"
                )

                # Create a dummy report. We must run all tests that did not
                # pass last time: with no result, all tests should run.
                self.old_report_index = ReportIndex(self.output_dir)

        # Rotate results directories if requested. In both cases, make sure the
        # new results dir is clean.
        if args.rotate_output_dirs:
            if os.path.isdir(old_output_dir):
                rm(old_output_dir, recursive=True)
            if os.path.isdir(self.output_dir):
                mv(self.output_dir, old_output_dir)
        elif os.path.isdir(self.output_dir):
            rm(self.output_dir, recursive=True)
        mkdir(self.output_dir)

        # Remember about the old output directory only if it exists and does
        # contain results. If not, this info will be unused at best, or lead to
        # incorrect behavior.
        self.old_output_dir = None
        if os.path.exists(old_output_dir) and os.path.exists(
            os.path.join(old_output_dir, ReportIndex.INDEX_FILENAME)
        ):
            self.old_output_dir = old_output_dir

        if args.dump_environ:
            dump_environ(os.path.join(self.output_dir, "environ.sh"), self.env)

    def run_standard_mainloop(self, dag: DAG) -> None:
        """Run the main loop to execute test fragments in threads."""
        assert self.main.args is not None

        from e3.job import Job
        from e3.testsuite.fragment import FragmentData, ThreadTestFragment

        def job_factory(
            uid: str,
            data: Any,
            predecessors: FrozenSet[str],
            notify_end: Callable[[str], None],
        ) -> ThreadTestFragment:
            """Turn a DAG item into a ThreadTestFragment instance."""
            assert isinstance(data, FragmentData)

            # When passing return values from predecessors, remove current test
            # name from the keys to ease referencing by user (the short
            # fragment name can then be used by user without knowing the full
            # node id).
            key_prefix = data.driver.test_name + "."
            key_prefix_len = len(key_prefix)

            def filter_key(k: str) -> str:
                if k.startswith(key_prefix):
                    return k[key_prefix_len:]
                else:
                    return k

            return ThreadTestFragment(
                uid,
                data.driver,
                data.callback,
                {filter_key(k): self.return_values[k] for k in predecessors},
                notify_end,
                self.running_status,
            )

        def collect_result(job: Job) -> bool:
            """Collect test results from the given fragment."""
            assert isinstance(job, ThreadTestFragment)
            self.return_values[job.uid] = job.return_value
            self.collect_result(job)

            # In the e3.job.scheduler API, collect returning "True" means
            # "requeue the job". We never want to do that.
            return False

        # Create a scheduler to run all fragments for the testsuite main loop

        jobs = self.main.args.jobs
        if self.main.args.jobs <= 0:
            jobs = os.cpu_count() or 1

        scheduler = Scheduler(
            job_provider=job_factory,
            tokens=jobs,
            collect=collect_result,
        )

        # Finally run the tests
        scheduler.run(dag)

    def run_multiprocess_mainloop(self, dag: DAG) -> None:
        """Run the main loop to execute test fragments in subprocesses."""
        assert self.main.args is not None

        from e3.testsuite.fragment import FragmentData, ProcessTestFragment
        from e3.testsuite.multiprocess_scheduler import MultiprocessScheduler

        def job_factory(
            uid: str, data: FragmentData, slot: int
        ) -> ProcessTestFragment:
            """Turn a DAG item into a ProcessTestFragment instance."""
            assert data.callback_by_name
            return ProcessTestFragment(
                uid,
                data.driver,
                data.name,
                slot,
                self.running_status,
                self.env,
            )

        def collect_result(job: ProcessTestFragment) -> None:
            """Collect test results from the given fragment."""
            job.collect_result()
            self.collect_result(job)

        scheduler: MultiprocessScheduler[FragmentData, ProcessTestFragment] = (
            MultiprocessScheduler(
                dag, job_factory, collect_result, jobs=self.main.args.jobs
            )
        )

        # Finally run the tests
        scheduler.run()

    # Unlike the previous methods, the following ones are supposed to be
    # overriden.

    @property
    def tests_subdir(self) -> str:
        """
        Return the subdirectory in which tests are looked for.

        The returned directory name is considered relative to the root
        testsuite directory (self.root_dir).
        """
        raise NotImplementedError

    @property
    def test_driver_map(self) -> Dict[str, Type[TestDriver]]:
        """Return a map from test driver names to TestDriver subclasses.

        Test finders will be able to use this map to fetch the test drivers
        referenced in testcases.
        """
        raise NotImplementedError

    @property
    def default_driver(self) -> Optional[str]:
        """Return the name of the default driver for testcases.

        When tests do not query a specific driver, the one associated to this
        name is used instead. If this property returns None, all tests are
        required to query a driver.
        """
        raise NotImplementedError

    def test_name(self, test_dir: str) -> str:
        """Compute the test name given a testcase spec.

        This function can be overridden. By default it uses the name of the
        test directory. Note that the test name should be a valid filename (not
        dir seprators, or special characters such as ``:``, ...).
        """
        raise NotImplementedError

    @property
    def test_finders(self) -> List[TestFinder]:
        """Return test finders to probe tests directories."""
        raise NotImplementedError

    def add_options(self, parser: argparse.ArgumentParser) -> None:
        """Add testsuite specific switches.

        Subclasses can override this method to add their own testsuite
        command-line options.

        :param parser: Parser for command-line arguments. See
            <https://docs.python.org/3/library/argparse.html> for usage.
        """
        raise NotImplementedError

    def set_up(self) -> None:
        """Execute operations before running the testsuite.

        Before running this, command-line arguments were parsed. After this
        returns, the testsuite will look for testcases.

        By default, this does nothing. Overriding this method allows testsuites
        to prepare the execution of the testsuite depending on their needs. For
        instance:

        * process testsuite-specific options;
        * initialize environment variables;
        * adjust self.env (object forwarded to test drivers).
        """
        raise NotImplementedError

    def tear_down(self) -> None:
        """Execute operation when finalizing the testsuite.

        By default, this cleans the working (temporary) directory in which the
        tests were run.
        """
        raise NotImplementedError

    def write_comment_file(self, comment_file: IO[str]) -> None:
        """Write the comment file's content.

        :param comment_file: File descriptor for the comment file.  Overriding
            methods should only call its "write" method (or print to it).
        """
        raise NotImplementedError

    @property
    def default_max_consecutive_failures(self) -> int:
        """Return the default maximum number of consecutive failures.

        In some cases, aborting the testsuite when there are just too many
        failures saves time and costs: the software to test/environment is too
        broken, there is no point to continue running the testsuite.

        This property must return the number of test failures (FAIL or ERROR)
        that trigger the abortion of the testuite. If zero, this behavior is
        disabled.
        """
        raise NotImplementedError

    @property
    def default_failure_exit_code(self) -> int:
        """Return the default exit code when at least one test fails."""
        raise NotImplementedError

    @property
    def auto_generate_text_report(self) -> bool:
        """Return whether to automatically generate a text report.

        This is disabled by default (and controlled by the
        --generate-text-report command-line option) because the generation of
        this report can add non-trivial overhead depending on results.
        """
        raise NotImplementedError

    def adjust_dag_dependencies(self, dag: DAG) -> None:
        """Adjust dependencies in the DAG of all test fragments.

        :param dag: DAG to adjust.
        :param fragments: Set of all fragments added so far to the DAG.
        """
        raise NotImplementedError

    @property
    def multiprocessing_supported(self) -> bool:
        """Return whether running test fragments in subprocesses is supported.

        When multiprocessing is enabled (see the "use_multiprocessing"
        attribute), test fragments are executed in a separate process, and the
        propagation of their return values is disabled (FragmentData's
        "previous_values" argument is always an empty dict).

        This means that multiprocessing can work only if test drivers and all
        code used by test fragments can be imported by subprocesses (for
        instance, class defined in the testsuite entry point are unavailable)
        and if test drivers don't use the "previous_values" mechanism.

        Testsuite authors can use the "--force-multiprocessing" testsuite
        option to check if this works.
        """
        raise NotImplementedError


class Testsuite(TestsuiteCore):
    """Testsuite class.

    When implementing a new testsuite you should create a class that
    inherit from this class.
    """

    @property
    def tests_subdir(self) -> str:
        return "."

    @property
    def test_driver_map(self) -> Dict[str, Type[TestDriver]]:
        raise NotImplementedError

    @property
    def default_driver(self) -> Optional[str]:
        return None

    def test_name(self, test_dir: str) -> str:
        # Start with a relative directory name from the tests subdirectory
        result = os.path.relpath(test_dir, self.test_dir)

        # We want to support running tests outside of the test directory, so
        # strip leading "..".
        pattern = os.path.pardir + os.path.sep
        while result.startswith(pattern):
            result = result[len(pattern) :]

        # Run some name canonicalization and replace directory separators with
        # double underscores.
        return result.replace("\\", "/").rstrip("/").replace("/", "__")

    @property
    def test_finders(self) -> List[TestFinder]:
        return [YAMLTestFinder()]

    def add_options(self, parser: argparse.ArgumentParser) -> None:
        pass

    def set_up(self) -> None:
        pass

    def tear_down(self) -> None:
        assert self.main.args

        if self.env.cleanup_mode == CleanupMode.ALL:
            rm(self.working_dir, True)

    def write_comment_file(self, comment_file: IO[str]) -> None:
        # Sensible default: just write the command line used to run the
        # testsuite. Testsuites can override or reuse this.
        quoted_cmdline = " ".join(quote_arg(arg) for arg in sys.argv)
        comment_file.write(f"Testsuite options:\n  {quoted_cmdline}\n")

    @property
    def default_max_consecutive_failures(self) -> int:
        return 0

    @property
    def default_failure_exit_code(self) -> int:
        return 1

    @property
    def auto_generate_text_report(self) -> bool:
        return False

    def adjust_dag_dependencies(self, dag: DAG) -> None:
        pass

    @property
    def multiprocessing_supported(self) -> bool:
        return False

    def compute_use_multiprocessing(self) -> bool:
        assert self.main.args

        # If multiprocessing is explicitly requested, just enable it
        if self.main.args.force_multiprocessing:
            return True

        # In practice, we noticed that running at most 16 jobs in parallel
        # creates neglectible GIL contention.
        if not self.multiprocessing_supported or self.main.args.jobs <= 16:
            return False

        return True
